import asyncio
import functools
import inspect


def _curry_helper(
    is_coroutine, signature, f, args_so_far, kwargs_so_far, *args, **kwargs
):
    f_len_args = signature.parameters
    args_so_far += args
    kwargs_so_far = {**kwargs_so_far, **kwargs}
    len_so_far = len(args_so_far) + len(kwargs_so_far)
    if len_so_far > len(f_len_args):
        return f(*args_so_far)
    if len_so_far == len(f_len_args):
        return f(*args_so_far, **kwargs_so_far)
    if len_so_far + 1 == len(f_len_args) and is_coroutine:

        @functools.wraps(f)
        async def curry_inner_async(*args, **kwargs):
            return await f(*(args_so_far + args), **{**kwargs_so_far, **kwargs})

        curry_inner_async.__signature__ = _update_signature(
            args_so_far,
            kwargs_so_far,
            signature,
        )
        return curry_inner_async

    @functools.wraps(f)
    def curry_inner(*args, **kwargs):
        return _curry_helper(
            is_coroutine, signature, f, args_so_far, kwargs_so_far, *args, **kwargs
        )

    curry_inner.__signature__ = _update_signature(
        args_so_far,
        kwargs_so_far,
        signature,
    )
    return curry_inner


def _update_signature(args_so_far, kwargs_so_far, signature):
    bounded = signature.bind_partial(
        *args_so_far,
        **kwargs_so_far,
    )
    new_signature = signature.replace(
        parameters=[
            v for k, v in signature.parameters.items() if k not in bounded.arguments
        ],
    )
    return new_signature


def _infer_defaults(params):
    kwargs = {}
    for p in params.values():
        if p.default != p.empty:
            kwargs[p.name] = p.default
    return kwargs


def curry(f):
    """Make a function handle partial input, returning a function that expects the rest.

    Warning: uses `inspect` which is slow, so avoid calling this inside a loop.

    >>> def addition(a, b): return a + b
    ... add_3 = gamla.curry(addition)(3)
    ... add_3(7)
    10

    Can also be used as a decorator:
    >>>@gamla.curry
    ... def addition(a, b):
    ...    return a + b

    In case the function is async, the function becomes synchronous until the last argument.
    Although this is not always what you want, it fits the majority of cases.

    >>> @gamla.curry
    ... async def async_addition(a, b):
    ...    await asyncio.sleep(0.1)
    ...    return a + b

    >>> add_3 = async_addition(3)

    >>> await add_3(7)  # Note that `await` is needed here. Must be done inside an async scope.

    The variables can be given with keywords, but mixing keyword and call by order might have unexpected results.
    """
    signature = inspect.signature(f)
    f_len_args = signature.parameters
    assert (
        len(f_len_args) > 1
    ), f"Curry function must have at least 2 parameters, {f} has {len(f_len_args)}"
    defaults = _infer_defaults(f_len_args)
    is_coroutine = asyncio.iscoroutinefunction(f)

    @functools.wraps(f)
    def indirection(*args, **kwargs):
        return _curry_helper(is_coroutine, signature, f, (), defaults, *args, **kwargs)

    return indirection



import typing
from collections import abc
from typing import Any, Callable, Optional, Tuple, TypeVar, Union

from gamla import construct, operator
from gamla.optimized import sync


def _handle_union_on_left(type1, type2):
    return sync.pipe(
        type1,
        typing.get_args,
        sync.allmap(lambda t: _is_subtype((t, type2))),
    )


def _handle_union_on_right(type1, type2):
    return sync.pipe(
        type2,
        typing.get_args,
        sync.anymap(lambda t: _is_subtype((type1, t))),
    )


_origin_equals = sync.compose_left(operator.equals, sync.before(typing.get_origin))

_handle_union = sync.case_dict(
    {
        sync.compose_left(operator.head, _origin_equals(Union)): sync.star(
            _handle_union_on_left,
        ),
        sync.compose_left(operator.second, _origin_equals(Union)): sync.star(
            _handle_union_on_right,
        ),
    },
)


def _rewrite_optional(x):
    return Union[None, typing.get_args(x)]


def _forward_ref(x):
    def forward_ref(*args, **kwargs):
        return x()(*args, **kwargs)

    return forward_ref


def _iterable_to_union(it):
    it = tuple(it)
    assert it
    if len(it) == 1:
        return it[0]
    return Union[it[0], _iterable_to_union(it[1:])]


_rewrite_typevar = sync.compose_left(
    operator.attrgetter("__constraints__"),
    sync.ternary(operator.empty, construct.just(Any), _iterable_to_union),
)

_handle_generics = sync.alljuxt(
    sync.compose_left(sync.map(typing.get_origin), sync.star(issubclass)),
    sync.compose_left(
        sync.map(typing.get_args),
        sync.star(zip),
        sync.allmap(_forward_ref(lambda: _is_subtype)),
    ),
)


def _handle_callable(c1, c2):
    args1 = typing.get_args(c1)
    args2 = typing.get_args(c2)
    if not args1 and args2:
        return False
    if not args2:
        return True
    input1, output1 = args1
    input2, output2 = args2
    return is_subtype(output1, output2) and (
        Ellipsis in [input1, input2]
        or len(input1) == len(input2)
        and sync.pipe([input1, input2], sync.star(zip), sync.allmap(_is_subtype))
    )


_is_subtype: Callable[[Tuple[Any, Any]], bool] = sync.compose_left(
    sync.map(
        sync.compose_left(
            sync.when(_origin_equals(Optional), _rewrite_optional),
            sync.when(operator.is_instance(TypeVar), _rewrite_typevar),
        ),
    ),
    tuple,
    sync.case_dict(
        {
            sync.allmap(_origin_equals(abc.Callable)): sync.star(_handle_callable),
            operator.inside(Any): sync.compose_left(
                operator.second,
                operator.equals(Any),
            ),
            sync.anymap(_origin_equals(Union)): _handle_union,
            sync.allmap(typing.get_origin): _handle_generics,
            operator.inside(Ellipsis): sync.allmap(operator.equals(Ellipsis)),
            sync.complement(sync.anymap(typing.get_origin)): sync.star(issubclass),
            construct.just(True): construct.just(False),
        },
    ),
)

#: Given two typings, checks if the second is a superset of the first.
is_subtype = sync.compose_left(operator.pack, _is_subtype)


_RETURN_TYPING = "return"


def composable(destination: Callable, origin: Callable, key: Optional[str]) -> bool:
    """Checks if `destination` can be composed after `source`, considering their typing."""
    s = typing.get_type_hints(origin)
    d = typing.get_type_hints(destination)
    if _RETURN_TYPING not in s:
        return True
    if key:
        if key not in d:
            return True
        d = d[key]
    else:
        if _RETURN_TYPING in d:
            del d[_RETURN_TYPING]
        if not d:
            return True
        if len(d) != 1:
            return False
        d = operator.head(d.values())
    return is_subtype(s[_RETURN_TYPING], d)



import pickle

from gamla import data, functional_generic, operator


def test_freeze_deep():
    original = {"1": {"2": "345", "some-string": ["hello"]}}
    frozen = data.freeze_deep(original)
    # Check values are intact.
    assert frozen == data.frozendict(
        {"1": data.frozendict({"2": "345", "some-string": ("hello",)})},
    )
    # Check hashability.
    {data.freeze_deep(original)}


def test_freeze_deep_idempotent():
    assert data.freeze_deep(data.freeze_deep({"a": 1})) == data.freeze_deep({"a": 1})


def test_frozendict_serializable():
    fd = data.frozendict({"a": "something", "b": 1})
    fd_str = pickle.dumps(fd)

    fd_clone = pickle.loads(fd_str)

    assert fd == fd_clone


def test_explode():
    assert functional_generic.pipe(
        [
            "x",
            [
                "y1",
                "y2",
                "y3",
            ],
            "z",
        ],
        data.explode(1),
        tuple,
    ) == (
        ("x", "y1", "z"),
        ("x", "y2", "z"),
        ("x", "y3", "z"),
    )


def test_transform_if_not_none():
    assert (
        data.transform_if_not_none(
            functional_generic.compose_left(
                functional_generic.when(
                    operator.is_instance(str),
                    lambda x: x.casefold(),
                ),
            ),
            "Some Title",
        )
        == "some title"
    )



from typing import Any, AsyncGenerator, Callable, Hashable, Text, Tuple

from gamla import construct, currying, functional_generic, operator


@currying.curry
async def agraph_traverse(
    source: Any,
    aget_neighbors: Callable[[Any], AsyncGenerator],
    key: Callable = operator.identity,
) -> AsyncGenerator:
    """Gets a graph and a function to get a node's neighbours,
    BFS over it from a single source node, returns an AsyncGenerator of unique nodes.

    >>> g = {'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []}
    >>> async def get_item(x):
    >>>     for a in g.get(x):
    >>>         yield a
    >>> async def to_list(ag):
    >>>     return [i async for i in ag]
    >>> gamla.run_sync(to_list(gamla.agraph_traverse('1', get_item)))
    ['1', '2', '3', '4']"""
    async for s in agraph_traverse_many(
        [source],
        aget_neighbors=aget_neighbors,
        key=key,
    ):
        yield s


@currying.curry
async def agraph_traverse_many(
    sources: Any,
    aget_neighbors: Callable[[Any], AsyncGenerator],
    key: Callable = operator.identity,
) -> AsyncGenerator[Any, None]:
    """BFS over a graph, yielding unique nodes.
    Use when `aget_neighbors` returns an AsyncGenerator.

    >>> g = {'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []}
    >>> async def get_item(x):
    >>>     for a in g.get(x):
    >>>         yield a
    >>> async def to_list(ag):
    >>>     return [i async for i in ag]
    >>> gamla.run_sync(to_list(gamla.agraph_traverse_many(['1', '3'], get_item)))
    ['3', '1', '4', '2']

    Note: `aget_neighbors` must return elements without duplicates."""
    queue = [*sources]
    seen = set(map(key, queue))

    while queue:
        current = queue.pop()
        yield current
        async for node in aget_neighbors(current):
            if key(node) not in seen:
                seen.add(key(node))
                queue = [node] + queue


@currying.curry
async def atraverse_graph_by_radius(
    source: Any,
    aget_neighbors: Callable[[Any], AsyncGenerator],
    radius: int,
    key: Callable = operator.identity,
) -> AsyncGenerator[Any, None]:
    """Gets a graph and a function to get a node's neighbours,
    BFS over it from a single source node, returns an AsyncGenerator of unique nodes.
    Does not traverse farther from given `radius`

    >>> g = {'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []}
    >>> async def get_item(x):
    >>>     for a in g.get(x):
    >>>         yield a
    >>> async def to_list(ag):
    >>>     return [i async for i in ag]
    >>> gamla.run_sync(to_list(gamla.atraverse_graph_by_radius('1', get_item, 1)))
    ['1', '2', '3']"""

    async def get_neighbors_limiting_radius(
        current_and_distance: Tuple[Text, int],
    ) -> AsyncGenerator[Tuple[Any, int], None]:
        current, distance = current_and_distance
        if distance < radius:
            async for neighbor in aget_neighbors(current):
                yield neighbor, distance + 1

    async for s in agraph_traverse(
        source=(source, 0),
        aget_neighbors=get_neighbors_limiting_radius,
        key=functional_generic.compose_left(operator.head, key),
    ):
        yield operator.head(s)


class _IgnoreChild:
    pass


@currying.curry
async def reduce_graph_async(
    reducer: Callable[[Tuple[Any, ...], Hashable], Any],
    get_neighbors: Callable,
    remember: Callable[[Hashable], None],
    is_seen: Callable[[Hashable], bool],
    current: Hashable,
):
    """Reduces a graph from some starting point using async functions.

    >>> set_instance = set()
    >>> await reduce_graph_async(
    ...     lambda children, current: sum(children) + current,
    ...     functional_generic.compose_left(
    ...         dict_utils.dict_to_getter_with_default(
    ...             (),
    ...             {1: (1, 2, 3, 5), 2: (4,), 3: (1, 2)}),
    ...         async_functions.to_awaitable,
    ...     ),
    ...     set_instance.add,
    ...     contains(set_instance)
    ...     1,
    ... )
    15"""
    if is_seen(current):
        return _IgnoreChild()
    # Since we may reach a node from two different branches, at the same time,
    # we have to broadcast to the other branch that we've reached this node, this
    # can't be done in an immutable fashion.
    remember(current)
    return await functional_generic.pipe(
        current,
        get_neighbors,
        functional_generic.curried_map(
            reduce_graph_async(reducer, get_neighbors, remember, is_seen),
        ),
        functional_generic.remove(operator.is_instance(_IgnoreChild)),
        tuple,
        functional_generic.pair_right(construct.just(current)),
        functional_generic.star(reducer),
    )



# flake8: noqa

from .apply_utils import *
from .construct import *
from .currying import *
from .data import *
from .debug_utils import *
from .dict_utils import *
from .excepts_decorator import *
from .functional import *
from .functional_async import *
from .functional_generic import *
from .graph import *
from .graph_async import *
from .higher_order import *
from .io_utils import *
from .operator import *
from .optimized import async_functions, sync
from .string_utils import *
from .tree import *
from .type_safety import *
from .url_utils import *

check = sync.check
map = functional_generic.curried_map
filter = functional_generic.curried_filter

to_awaitable = async_functions.to_awaitable

NoConditionMatched = sync.NoConditionMatched



import asyncio
import random

import pytest

from gamla import functional_generic, io_utils


async def test_batch_decorator():
    times_f_called = 0

    @io_utils.batch_calls(100)
    async def slow_identity(inputs):
        nonlocal times_f_called
        times_f_called += 1
        await asyncio.sleep(random.random() * 0.1 * len(inputs))
        return inputs

    inputs = tuple(range(100))
    results = await functional_generic.pipe(
        inputs,
        functional_generic.curried_map(slow_identity),
        tuple,
    )
    assert results == inputs
    assert times_f_called < 5


async def test_batch_decorator_errors():
    times_f_called = 0

    @io_utils.batch_calls(100)
    async def slow_identity_with_errors(inputs):
        nonlocal times_f_called
        times_f_called += 1
        await asyncio.sleep(random.random() * 0.1 * len(inputs))
        if times_f_called > 1:
            raise ValueError
        return inputs

    assert (
        await functional_generic.pipe(
            (1,),
            functional_generic.curried_map(slow_identity_with_errors),
            tuple,
        )
    ) == (1,)

    with pytest.raises(ValueError):
        await functional_generic.pipe(
            (1, 2, 3),
            functional_generic.curried_map(slow_identity_with_errors),
            tuple,
        )

    assert times_f_called == 2


async def test_batch_decorator_max_size():
    times_f_called = 0

    @io_utils.batch_calls(10)
    async def slow_identity(inputs):
        nonlocal times_f_called
        times_f_called += 1
        await asyncio.sleep(random.random() * 0.1 * len(inputs))
        return inputs

    inputs = tuple(range(100))
    results = await functional_generic.pipe(
        inputs,
        functional_generic.curried_map(slow_identity),
        tuple,
    )
    assert results == inputs
    assert times_f_called == 10


async def test_with_and_without_deduper():
    inputs = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 3)
    called = []

    async def identity_with_spying_and_delay(x):
        await asyncio.sleep(0.4)
        called.append(x)
        return x

    # Without.
    assert inputs == tuple(
        await functional_generic.curried_map(identity_with_spying_and_delay)(inputs),
    )

    assert len(called) == len(inputs)

    called[:] = []

    # With.
    assert inputs == tuple(
        await functional_generic.curried_map(
            io_utils.queue_identical_calls(identity_with_spying_and_delay),
        )(
            inputs,
        ),
    )

    unique = frozenset(inputs)
    assert len(called) == len(unique)
    assert frozenset(called) == unique


async def test_retry():
    class SomeError(Exception):
        pass

    succeeds_in_n_retries = 3

    async def f(x, y):
        nonlocal succeeds_in_n_retries
        if succeeds_in_n_retries == 0:
            return x + y
        succeeds_in_n_retries -= 1
        raise SomeError

    assert await io_utils.retry(SomeError, 3, 0, f)(3, 2) == 5


async def test_retry_raises():
    class SomeError(Exception):
        pass

    succeeds_in_n_retries = 3

    async def f(x, y):
        nonlocal succeeds_in_n_retries
        if succeeds_in_n_retries == 0:
            return x + y
        succeeds_in_n_retries -= 1
        raise SomeError

    with pytest.raises(SomeError):
        await io_utils.retry(SomeError, 2, 0, f)(3, 2)


async def test_throtle():
    factor = 0

    @io_utils.throttle(1)
    async def multiply_with_delay(x):
        nonlocal factor
        factor = factor + 1
        await asyncio.sleep(0.01)
        return x * factor

    assert await functional_generic.pipe(
        (1, 2, 3),
        functional_generic.curried_map(multiply_with_delay),
        tuple,
    ) == (1, 4, 9)


async def test_make_throttler():
    total = 0
    throttle = io_utils.make_throttler(1)

    @throttle
    async def inc():
        nonlocal total
        total = total + 1
        await asyncio.sleep(0.01)
        assert total <= 1

    @throttle
    async def dec():
        await asyncio.sleep(0.01)
        nonlocal total
        total = total - 1
        assert total >= 0

    await asyncio.gather(inc(), dec(), inc(), dec())
    assert total == 0



import re
from typing import Text


def replace_in_text(old: Text, new: Text):
    """Return a copy of the string with all occurrences of substring old replaced by new
    >>> txt = "hello world"
    >>> replace_in_text("world", "Jhon")(txt)
    'hello Jhon'
    """

    def replace_in_text(txt: Text):
        return txt.replace(old, new)

    return replace_in_text


def split_text(sep: Text):
    """Return a list of the words in the string, using sep as the delimiter string

    >>> txt = "hello world"
    >>> split_text(" ")(txt)
    ['hello', 'world']
    """

    def split_text(txt: Text):
        return txt.split(sep)

    return split_text


def capitalize(text: Text):
    """Capitalize only the first letter of a given string, unlike
    str.capitalize that transforms all characters after the first
    character to lowercase.

    >>>capitalize("my name is John Smith")
    'My name is John Smith'
    """
    return text[0].upper() + text[1:]


def regex_match(pattern):
    """A curried regex match. Gets a pattern and returns a function that expects a text to match the pattern with.

    >>> regex_match(r"phone:(\d*)")("phone:1234567").group(1)  # noqa: W605
    '1234567'

    """

    def regex_match_inner(text):
        return re.match(pattern, text)

    return regex_match_inner
